"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RuleEngineImpl = void 0;
const dependency_graph_1 = require("../dependency-graph/dependency-graph");
const builder_1 = require("./builder/builder");
const validation_type_1 = require("../validators/validation-type");
const rule_set_1 = require("./rule-set/rule-set");
const validation_result_1 = require("../validators/validation-result");
class RuleEngineImpl {
    constructor(options) {
        this.propertyMap = {};
        this.dependencyGraph = new dependency_graph_1.DependencyGraph();
        this.validations = new WeakMap();
        this.dataLinks = new Map();
        this.snapshots = new Map();
        this.builder = new builder_1.Builder(options, this, this.dependencyGraph, this.propertyMap);
        this.version = options.version;
        this.dataMigrators = options.dataMigrators;
    }
    get properties() {
        return Object.values(this.propertyMap);
    }
    getVersion() {
        return this.version;
    }
    isCompatible(data) {
        return this.getVersion().compatibleWith(data.rulesVersion);
    }
    getBuilder() {
        return this.builder;
    }
    getPropertyById(id) {
        return this.propertyMap[id];
    }
    validateAllProperties() {
        return __awaiter(this, void 0, void 0, function* () {
            yield Promise.all(Object.values(this.propertyMap).map(prop => prop.validate()));
            return this.getValidationMessages();
        });
    }
    setValidationMessages(validationMessages) {
        this.clearValidationResult();
        const validationMessagesMap = validationMessages instanceof validation_result_1.ValidationResult
            ? validationMessages.getValidationMessagesMap() : validationMessages;
        const unknownIds = [];
        Object.keys(validationMessagesMap).forEach(key => {
            if (this.propertyMap[key] !== undefined) {
                this.propertyMap[key].setValidationMessages(validationMessagesMap[key]);
            }
            else {
                unknownIds.push(key);
            }
        });
        return unknownIds;
    }
    getValidationMessages() {
        const map = Object.keys(this.propertyMap).reduce((res, cur) => {
            const messages = this.propertyMap[cur].getValidationMessages();
            if (messages.length > 0) {
                res[cur] = messages;
            }
            return res;
        }, {});
        return new validation_result_1.ValidationResult(map, (id) => this.getPropertyById(id));
    }
    clearValidationResult() {
        this.properties.forEach(prop => prop.clearValidationResult());
    }
    exportData(onlySelectedProperties) {
        const rulesVersion = this.getVersion().version;
        const propsToExport = onlySelectedProperties ? onlySelectedProperties : this.properties.filter(prop => !this.dependencyGraph.isOwnedProperty(prop.id));
        const data = propsToExport.reduce((res, prop) => {
            res[prop.id] = prop.exportData();
            return res;
        }, {});
        return {
            rulesVersion,
            data
        };
    }
    importData(data) {
        var _a;
        (_a = this.dataMigrators) === null || _a === void 0 ? void 0 : _a.forEach(migrator => {
            if (migrator.acceptsVersion(data.rulesVersion) != null) {
                data = {
                    rulesVersion: migrator.newVersion,
                    data: migrator.migrate(data)
                };
            }
        });
        if (!this.isCompatible(data)) {
            throw new Error(`Version "${data.rulesVersion}" of imported data is not compatible with current version "${this.getVersion().version}" and there are missing porper data migrators`);
        }
        this.setToInitialState();
        Object.entries(data.data).forEach(([propertyId, propData]) => {
            const prop = this.getPropertyById(propertyId);
            prop === null || prop === void 0 ? void 0 : prop.importData(propData);
        });
    }
    setToInitialState() {
        this.properties.forEach(property => property.setToInitialState());
    }
    takeSnapShot(key = 'default') {
        const snap = {
            key,
            data: this.exportData()
        };
        this.snapshots.set(key, snap);
        return snap;
    }
    restoreSnapShot(key = 'default') {
        const snap = this.snapshots.get(key);
        if (snap) {
            this.importData(snap.data);
        }
    }
    linkPropertyData(propertyA, propertyB) {
        const dataLinkKey = this.createDataLinkKey(propertyA, propertyB);
        if (this.dataLinks.has(dataLinkKey)) {
            return;
        }
        const listenerA = {
            updated: () => propertyA.transferData(propertyB)
        };
        const listenerB = {
            updated: () => propertyB.transferData(propertyA)
        };
        propertyA.transferData(propertyB);
        const refA = propertyA.registerValueChangedListener(listenerA);
        const refB = propertyB.registerValueChangedListener(listenerB);
        this.dataLinks.set(dataLinkKey, [refA, refB]);
    }
    unlinkPropertyData(propertyA, propertyB) {
        const dataLinkKey = this.createDataLinkKey(propertyA, propertyB);
        const listeners = this.dataLinks.get(dataLinkKey);
        if (listeners) {
            const [refA, refB] = listeners;
            propertyA.deregisterValueChangedListener(refA);
            propertyB.deregisterValueChangedListener(refB);
            this.dataLinks.delete(dataLinkKey);
        }
    }
    createDataLinkKey(propertyA, propertyB) {
        return propertyA.id < propertyB.id
            ? `${propertyA.id}<->${propertyB.id}`
            : `${propertyB.id}<->${propertyA.id}`;
    }
    needsAnUpdate(mightHaveChanged) {
        this.dependencyGraph.traverseDepthFirst(mightHaveChanged.id, property => {
            property.needsAnUpdate(false);
        }, dependency => !!dependency.options.value);
    }
    updateValue(property) {
        var _a;
        const asyncUpdates = (_a = this.dependencyGraph.getAsyncDependencies(property.id)) === null || _a === void 0 ? void 0 : _a.map(asyncDep => this.updateValue(asyncDep));
        if ((asyncUpdates === null || asyncUpdates === void 0 ? void 0 : asyncUpdates.length) && asyncUpdates.some(update => !!update)) {
            return Promise.all(asyncUpdates).then(() => this.updateValueInternal(property)).catch((err) => {
                property.errorWhileUpdating(err);
            });
        }
        else {
            return this.updateValueInternal(property);
        }
    }
    updateValueInternal(property) {
        try {
            const updatePromise = property.internallyUpdate();
            if (updatePromise) {
                return updatePromise.then(() => {
                    this.hasBeenUpdated(property);
                }).catch(err => property.errorWhileUpdating(err));
            }
            else {
                this.hasBeenUpdated(property);
            }
        }
        catch (err) {
            property.errorWhileUpdating(err);
        }
    }
    hasBeenUpdated(property) {
        property.hasBeenUpdated();
        this.dependencyGraph.traverseDepthFirst(property.id, (prop, dep) => prop.dependencyHasBeenUpdated(dep), dependency => !dependency.options.value, true);
    }
    cancelValidationAndInvalidateResults(validators) {
        validators.forEach(v => {
            const vprocess = this.getValidationProcess(v);
            vprocess.isLastResultUpToDate = false;
            vprocess.isCancelled = true;
        });
    }
    validateValidatorInstances(validators) {
        return validators.map(validator => {
            const vprocess = this.getValidationProcess(validator);
            if (vprocess.isLastResultUpToDate) {
                return Promise.resolve(vprocess.lastValidationResult);
            }
            else if (vprocess.currentValidation && !vprocess.isCancelled) {
                return vprocess.currentValidation;
            }
            vprocess.isCancelled = false;
            const validation = validator.validate(...validator.validationArguments);
            if (validation instanceof Promise) {
                vprocess.currentValidation = validation.then(res => {
                    vprocess.currentValidation = undefined;
                    vprocess.lastValidationResult = res;
                    if (vprocess.isCancelled) {
                        return 'cancelled';
                    }
                    else {
                        vprocess.isLastResultUpToDate = true;
                        return res;
                    }
                }, err => {
                    vprocess.currentValidation = undefined;
                    return [{
                            text: JSON.stringify(err),
                            type: validation_type_1.ValidationType.ErrorThrownDuringValidation
                        }];
                });
                return vprocess.currentValidation;
            }
            else {
                vprocess.lastValidationResult = validation;
                vprocess.isLastResultUpToDate = true;
                return Promise.resolve(validation);
            }
        });
    }
    getValidationProcess(validator) {
        const vprocess = this.validations.get(validator);
        if (vprocess) {
            return vprocess;
        }
        else {
            const newProcess = {
                isLastResultUpToDate: false,
                isCancelled: false,
            };
            this.validations.set(validator, newProcess);
            return newProcess;
        }
    }
    removeOwnedProperties(ownerId, ownedProperties) {
        const removed = this.dependencyGraph.removeOwnedProperties(ownerId, ownedProperties);
        if (removed) {
            removed.forEach(id => {
                delete this.propertyMap[id];
                this.removeOwnedProperties(id);
            });
        }
    }
    defineRuleSet(initFcn) {
        return new rule_set_1.RuleSet(initFcn, this.builder);
    }
}
exports.RuleEngineImpl = RuleEngineImpl;
