"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AbstractPropertyImpl = void 0;
const logger_1 = require("../util/logger/logger");
const assertion_error_1 = require("../util/assertions/assertion-error");
const validation_result_1 = require("../validators/validation-result");
class AbstractPropertyImpl {
    constructor(updateHandler, backpressureConfig) {
        this.updateHandler = updateHandler;
        this.backpressureConfig = backpressureConfig;
        this.automaticallyUpdate = true;
        this.validationMessages = [];
    }
    awaitAsyncUpdate() {
        if (this.needsToRecompute !== false && this.automaticallyUpdate) {
            return this.updateHandler.updateValue(this);
        }
        else {
            return this.updatingPromise();
        }
    }
    updatingPromise() {
        if (this.currentRecomputing) {
            return new Promise((resolve, reject) => {
                const ul = { resolve, reject };
                if (!this.updatedListeners) {
                    this.updatedListeners = [ul];
                }
                else {
                    this.updatedListeners.push(ul);
                }
            });
        }
    }
    nextRecomputingCount() {
        var _a;
        this.recomputingCount = (((_a = this.recomputingCount) !== null && _a !== void 0 ? _a : 0) % Number.MAX_SAFE_INTEGER) + 1;
        return this.recomputingCount;
    }
    syncUpdateIfNeeded() {
        if (!this.isAsynchronous() && this.needsToRecompute !== false && this.automaticallyUpdate) {
            void this.updateHandler.updateValue(this);
        }
    }
    getTriggerListener() {
        if (!this.triggerListener) {
            this.triggerListener = {
                onTriggered: () => {
                    var _a;
                    if (this.needsToRecompute !== false) {
                        this.manuallyTriggered = true;
                        return (_a = this.updateHandler.updateValue(this)) !== null && _a !== void 0 ? _a : Promise.resolve();
                    }
                    else {
                        return Promise.resolve();
                    }
                }
            };
        }
        return this.triggerListener;
    }
    internallyUpdate() {
        if (this.needsToRecompute === false || (!this.automaticallyUpdate && !this.manuallyTriggered)) {
            return this.updatingPromise();
        }
        this.needsToRecompute = false;
        delete this.manuallyTriggered;
        if (this.isAsynchronous()) {
            if (this.currentRecomputing) {
                return this.handleBackpressure();
            }
            else {
                return this.handleAsyncUpdate(this.nextRecomputingCount(), false);
            }
        }
        else {
            this.internallySyncUpdate();
        }
    }
    handleBackpressure() {
        var _a;
        if (!this.backpressureConfig) {
            throw new assertion_error_1.AssertionError('AbstractPropertyImpl: backpressureConfig needs to be defined');
        }
        switch (this.backpressureConfig.type) {
            case 'switch':
                if (this.backpressureConfig.debounceTime) {
                    return this.handleAsyncUpdateWithDebounceTime(this.backpressureConfig.debounceTime);
                }
                else {
                    return this.handleAsyncUpdate(this.nextRecomputingCount(), true);
                }
            case 'skip':
            default:
                return (_a = this.updatingPromise()) !== null && _a !== void 0 ? _a : Promise.resolve();
        }
    }
    handleAsyncUpdateWithDebounceTime(debounceTime) {
        return __awaiter(this, void 0, void 0, function* () {
            const recomputingCount = this.nextRecomputingCount();
            yield Promise.race([
                this.currentRecomputing,
                new Promise(resolve => setTimeout(() => resolve(), debounceTime))
            ]);
            return this.handleAsyncUpdate(recomputingCount, true);
        });
    }
    handleAsyncUpdate(recomputingCount, wasAlreadyProcessing) {
        var _a, _b;
        if (recomputingCount !== this.recomputingCount) {
            logger_1.Logger.trace(() => `Property.handleAsyncUpdate ${this.id}: ${recomputingCount} is skipped (${this.recomputingCount})`);
            return (_a = this.updatingPromise()) !== null && _a !== void 0 ? _a : Promise.resolve();
        }
        logger_1.Logger.trace(() => `Property.handleAsyncUpdate ${this.id}: ${recomputingCount}`);
        const update = this.internallyAsyncUpdate();
        this.currentRecomputing = update.asyncPromise.then((value) => {
            var _a;
            logger_1.Logger.trace(() => `Property.handleAsyncUpdate ${this.id}: ${recomputingCount} / ${this.recomputingCount}`);
            if (recomputingCount === this.recomputingCount) {
                logger_1.Logger.trace(() => `Property.handleAsyncUpdate ${this.id}: updated`);
                this.currentRecomputing = undefined;
                update.resolve(value);
                (_a = this.updatedListeners) === null || _a === void 0 ? void 0 : _a.forEach(ul => ul.resolve());
                delete this.updatedListeners;
            }
            else if (this.currentRecomputing) {
                return this.currentRecomputing;
            }
        });
        if (this.isProcessing() && !wasAlreadyProcessing) {
            this.tellValueChangeListeners(listener => { var _a; return (_a = listener.startsAsyncUpdate) === null || _a === void 0 ? void 0 : _a.call(listener); });
        }
        return (_b = this.updatingPromise()) !== null && _b !== void 0 ? _b : Promise.resolve();
    }
    needsAnUpdate(notifyOthers, needsToRecompute) {
        logger_1.Logger.trace(() => `Property.needsAnUpdate ${this.id}`);
        this.needsToRecompute = needsToRecompute !== null && needsToRecompute !== void 0 ? needsToRecompute : true;
        this.cancelValidationAndInvalidateResults();
        if (notifyOthers !== false) {
            this.updateHandler.needsAnUpdate(this);
        }
        this.tellValueChangeListeners(listener => { var _a; return (_a = listener.needsAnUpdate) === null || _a === void 0 ? void 0 : _a.call(listener); });
    }
    hasBeenUpdated() {
        this.tellValueChangeListeners(listener => listener.updated());
    }
    errorWhileUpdating(error) {
        var _a;
        (_a = this.updatedListeners) === null || _a === void 0 ? void 0 : _a.forEach(ul => ul.reject(error));
        delete this.updatedListeners;
        this.tellValueChangeListeners(listener => { var _a; return (_a = listener.updateFailed) === null || _a === void 0 ? void 0 : _a.call(listener, error); });
    }
    dependencyHasBeenUpdated(dependency) {
        if (dependency.options.validation) {
            this.cancelValidationAndInvalidateResults();
        }
        this.tellValueChangeListeners(listener => { var _a; return (_a = listener.dependencyHasBeenUpdated) === null || _a === void 0 ? void 0 : _a.call(listener, dependency); });
    }
    addPropertyValidator(validator, dependencies) {
        if (!this.propertyValidators) {
            this.propertyValidators = [];
        }
        this.propertyValidators.push({
            validationArguments: [...dependencies],
            validate: validator
        });
    }
    addAsyncPropertyValidator(validator, dependencies) {
        if (!this.asyncPropertyValidators) {
            this.asyncPropertyValidators = [];
        }
        this.asyncPropertyValidators.push({
            validationArguments: [...dependencies],
            validate: validator
        });
    }
    getPropertyValidatorResults() {
        var _a, _b;
        return (_b = (_a = this.propertyValidators) === null || _a === void 0 ? void 0 : _a.reduce((res, validatorInstance) => {
            const msg = validatorInstance.validate(this, ...validatorInstance.validationArguments);
            if (msg) {
                res.push(msg);
            }
            return res;
        }, [])) !== null && _b !== void 0 ? _b : [];
    }
    getAsyncPropertyValidatorResults() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.asyncPropertyValidators) {
                const messages = yield Promise.all(this.asyncPropertyValidators.map(validatorInstance => validatorInstance.validate(this, ...validatorInstance.validationArguments)));
                return messages.filter(msg => msg !== undefined);
            }
            return [];
        });
    }
    addValidator(validator) {
        if (!this.crossValidators) {
            this.crossValidators = [];
        }
        this.crossValidators.push(validator);
    }
    validate() {
        return __awaiter(this, void 0, void 0, function* () {
            const updated = this.awaitAsyncUpdate();
            if (updated) {
                yield updated;
            }
            const validationMessages = this.getPropertyValidatorResults();
            this.updateValidationMessages(validationMessages);
            const asyncValidationMessagesPromise = this.asyncPropertyValidators ? this.getAsyncPropertyValidatorResults() : undefined;
            if (this.crossValidators) {
                const results = yield Promise.all(this.updateHandler.validateValidatorInstances(this.crossValidators));
                results.filter(res => res !== 'cancelled').forEach(result => {
                    if (result instanceof Array) {
                        validationMessages.push(...result);
                    }
                    else if (result instanceof Object && result[this.id]) {
                        validationMessages.push(...result[this.id]);
                    }
                });
            }
            if (asyncValidationMessagesPromise) {
                validationMessages.push(...yield asyncValidationMessagesPromise);
            }
            return this.getValidationMessages();
        });
    }
    validateRecursively() {
        return __awaiter(this, void 0, void 0, function* () {
            const validationMap = yield this.validateRecursivelyInternal();
            return new validation_result_1.ValidationResult(validationMap, (id) => this.updateHandler.getPropertyById(id));
        });
    }
    validateRecursivelyInternal() {
        return __awaiter(this, void 0, void 0, function* () {
            return {
                [this.id]: yield this.validate()
            };
        });
    }
    isValid() {
        return this.validationMessages.every(msg => msg.type.isValid);
    }
    getValidationMessages() {
        return this.validationMessages;
    }
    setValidationMessages(messages) {
        this.cancelValidationAndInvalidateResults();
        this.updateValidationMessages(messages);
    }
    clearValidationResult() {
        this.cancelValidationAndInvalidateResults();
        this.updateValidationMessages([]);
    }
    cancelValidationAndInvalidateResults() {
        if (this.crossValidators) {
            this.updateHandler.cancelValidationAndInvalidateResults(this.crossValidators);
        }
    }
    updateValidationMessages(validationMessages) {
        this.validationMessages = validationMessages;
        this.tellValueChangeListeners(listener => { var _a; return (_a = listener.validated) === null || _a === void 0 ? void 0 : _a.call(listener); });
    }
    registerValueChangedListener(changed) {
        var _a;
        if (!this.valueChangeListeners) {
            this.valueChangeListeners = [];
        }
        const id = (_a = this.nextValueChangeListenerId) !== null && _a !== void 0 ? _a : 0;
        this.nextValueChangeListenerId = id + 1;
        const ref = { id };
        this.valueChangeListeners.push([id, changed]);
        if (changed.needsAnUpdate && this.needsToRecompute !== false) {
            changed.needsAnUpdate();
        }
        return ref;
    }
    deregisterValueChangedListener(changed) {
        if (this.valueChangeListeners && changed !== undefined) {
            const index = this.valueChangeListeners.findIndex(e => e[0] === changed.id);
            if (index >= 0) {
                this.valueChangeListeners.splice(index, 1);
            }
        }
    }
    hasValueChangeListeners() {
        return this.valueChangeListeners !== undefined && this.valueChangeListeners.length > 0;
    }
    tellValueChangeListeners(fcn) {
        var _a;
        (_a = this.valueChangeListeners) === null || _a === void 0 ? void 0 : _a.forEach(e => fcn(e[1]));
    }
    setUpdateStrategy(strategy) {
        this.automaticallyUpdate = strategy === 'Automatic';
    }
    addUpdateTrigger(trigger, setUpdateStrategyManually = true) {
        if (setUpdateStrategyManually) {
            this.setUpdateStrategy("Manually");
        }
        trigger.registerTriggerListener(this.getTriggerListener());
    }
    removeUpdateTrigger(trigger) {
        if (this.triggerListener) {
            trigger.deregisterTriggerListener(this.triggerListener);
        }
    }
    setTouched(touched) {
        this.touched = touched;
    }
    isTouched() {
        var _a;
        return (_a = this.touched) !== null && _a !== void 0 ? _a : false;
    }
    defineLabel(label) {
        this.label = label;
    }
    getLabel() {
        var _a;
        return (_a = this.label) !== null && _a !== void 0 ? _a : this.id;
    }
    defineAttribute(attribute) {
        if (!this.attributeMap) {
            this.attributeMap = new Map();
        }
        this.attributeMap.set(attribute.id, attribute);
    }
    get(id) {
        var _a, _b;
        return (_b = (_a = this.attributeMap) === null || _a === void 0 ? void 0 : _a.get(id)) === null || _b === void 0 ? void 0 : _b.getValue();
    }
    defineVisibility(visibility) {
        this.visible = visibility;
    }
    isVisible() {
        var _a, _b;
        return (_b = (_a = this.visible) === null || _a === void 0 ? void 0 : _a.getValue()) !== null && _b !== void 0 ? _b : true;
    }
    transferData(toProperty) {
        toProperty.importData(this.exportData());
    }
    setToInitialState() {
        this.clearValidationResult();
        this.needsToRecompute = true;
        this.currentRecomputing = undefined;
        this.setDataToInitialState();
        this.touched = false;
    }
    removeOwnedProperties(properties) {
        this.updateHandler.removeOwnedProperties(this.id, properties === null || properties === void 0 ? void 0 : properties.map(prop => prop.id));
    }
}
exports.AbstractPropertyImpl = AbstractPropertyImpl;
