import { alwaysAssertThat } from "../../util/assertions/assertions";
import { A } from "../../attributes/attribute-id";
export class AbstractPropertyRuleBuilder {
    constructor(property, addDependencies, textInterpreters) {
        this.property = property;
        this.addDependencies = addDependencies;
        this.textInterpreters = textInterpreters;
    }
    addValidatorInternal(...dependencies) {
        this.addDependencies(dependencies, this.property, { validation: true });
        return validator => {
            this.property.addPropertyValidator(validator, dependencies);
            return this;
        };
    }
    addAsyncValidatorInternal(...dependencies) {
        this.addDependencies(dependencies, this.property, { validation: true });
        return validator => {
            this.property.addAsyncPropertyValidator(validator, dependencies);
            return this;
        };
    }
    set(attribudeId, value) {
        this.defineAttributeFunction(attribudeId, [], () => value);
        return this;
    }
    define(attribudeId, ...dependencies) {
        return (value) => {
            this.defineAttributeFunction(attribudeId, dependencies, () => value(this.property, ...dependencies));
            return this;
        };
    }
    defineAttributeFunction(attribudeId, dependencies, fcn) {
        alwaysAssertThat(!A[attribudeId.name], () => {
            const mapping = {
                Required: 'defineRequiredIfVisible',
                Visible: 'defineVisibility',
                Label: 'defineLabel',
                InfoText: 'defineInfoText'
            };
            return `Please use specialised function ${mapping[attribudeId.name]} to define ${attribudeId.name}`;
        });
        this.addDependencies(dependencies, this.property, { [attribudeId.name]: true });
        this.property.defineAttribute({
            id: attribudeId,
            dependencies: dependencies,
            getValue: fcn
        });
    }
    setVisibility(isVisible) {
        if (isVisible === false) {
            this.defineVisibilityFunction([], () => isVisible);
        }
        else {
            this.property.defineVisibility(undefined);
        }
        return this;
    }
    defineVisibility(...dependencies) {
        return (isVisible) => {
            this.defineVisibilityFunction(dependencies, () => isVisible(this.property, ...dependencies));
            return this;
        };
    }
    defineVisibilityFunction(dependencies, fcn) {
        this.addDependencies(dependencies, this.property, { visible: true });
        this.property.defineVisibility({
            id: A.Visible,
            dependencies: dependencies,
            getValue: fcn
        });
    }
    defineLabel(text, textInterpreter) {
        this.checkTextInterpreterAvailable(textInterpreter);
        this.property.defineLabel(this.interpreteText(text, textInterpreter));
        return this;
    }
    interpreteText(text, textInterpreter) {
        var _a;
        if (textInterpreter !== undefined) {
            const interpreter = this.textInterpreters[textInterpreter];
            return (_a = interpreter === null || interpreter === void 0 ? void 0 : interpreter.interpreteText(text)) !== null && _a !== void 0 ? _a : text;
        }
        return text;
    }
    checkTextInterpreterAvailable(textInterpreter) {
        alwaysAssertThat(textInterpreter === undefined || this.textInterpreters[textInterpreter] !== undefined, () => `Text interpreter '${textInterpreter}' not found. Please provide via builder options.`);
    }
    onUpdated(updated) {
        this.property.registerValueChangedListener({
            updated: () => updated(this.property)
        });
        return this;
    }
    setToInitialStateOnOtherPropertyChanged(...otherProperties) {
        alwaysAssertThat(!this.property.isReadOnly(), () => `${this.property.id}: Can only set properties to initial value, that are not read only`);
        otherProperties.forEach(prop => prop.registerValueChangedListener({
            updated: () => this.property.setToInitialState()
        }));
        return this;
    }
}
